/*
 *
 * $Id: lcd.c,v 1.1 2009/01/11 12:54:10 fvecoven Exp $
 *
 * LCD driver.
 */

#include <p18cxxx.h>
#include <delays.h>
#include "lcd.h"



void delay_ms(unsigned char n)
{
	unsigned char i;
	
	for (i = 0; i < n; i++) {
		Delay100TCYx(10);	
	}
}	

static unsigned char lcd_read(void)
{
	unsigned char tmp;
	unsigned char data;
	
	/* set data bits to input */
	LCD_DATA_TRIS |= 0x0F;
	
	/* read = 1 */
	LCD_RW = 1;
	
	Nop();
	Nop();
	Nop();
	Nop();
	Nop();
	Nop();
	Nop();
	Nop();
	
	/* clock high */
	LCD_E = 1;

	Nop();
	Nop();
	Nop();
	Nop();
	Nop();
	Nop();
	Nop();
	Nop();

	/* read high nibble */
	tmp = LCD_DATA;
	data = (tmp << 4);
	
	LCD_E = 0;
	
	Nop();
	Nop();
	Nop();
	Nop();
	Nop();
	Nop();
	Nop();
	Nop();
	
	LCD_E = 1;

	Nop();
	Nop();
	Nop();
	Nop();
	Nop();
	Nop();
	Nop();
	Nop();
	
	/* read low nibble */
	tmp = LCD_DATA;
	tmp &= 0x0F;
	data |= tmp;
	
	LCD_E = 0;
	
	Nop();
	Nop();
	Nop();
	Nop();
	Nop();
	Nop();
	Nop();
	Nop();
	
	return (data);
}


void lcd_busy(void)
{
	unsigned char data;
	
	do {
		LCD_RS = 0;
		data = lcd_read();
	} while ((data & 0x80) != 0);
}	

static void lcd_write_nibble(unsigned char nibble)
{
	/* set write mode */
	LCD_RW = 0;
	
	/* set output bits */
	LCD_DATA_TRIS &= 0xF0;
	
	Nop();
	Nop();
	Nop();
	Nop();
	
	/* clock high */
	LCD_E = 1;
	
	LCD_DATA &= 0xF0;
	LCD_DATA |= (nibble & 0x0F);
	
	Nop();
	Nop();
	Nop();
	Nop();
	
	/* clock low */
	LCD_E = 0;
	
	return;
}


static void lcd_write(unsigned char data)
{
	lcd_write_nibble(data >> 4);
	lcd_write_nibble(data & 0x0F);

	return;
}
	

void lcd_i_write(unsigned char data)
{
	lcd_busy();
	LCD_RS = 0;
	lcd_write(data);	
}	

void lcd_d_write(unsigned char data)
{
	lcd_busy();
	LCD_RS = 1;
	lcd_write(data);
}	

	
void lcd_init(void)
{
	/* PIC18F452 : port A = all digital except RA0 */
	ADCON1 = 0x0E;

	LCD_DATA = 0;
	
	LCD_E_DIR = 0;
	LCD_RS_DIR = 0;
	LCD_RW_DIR = 0;
	
	delay_ms(15);
	
	/* #1 control sequence */
	lcd_write_nibble(0b00000011);
	
	delay_ms(6);
	
	/* #2 control sequence */
	lcd_write_nibble(0b00000011);
	
	delay_ms(1);
	
	/* #3 control sequence */
	lcd_write_nibble(0b00000011);
	
	delay_ms(1);
	
	/* #4 set 4 bits */
	lcd_write_nibble(0b00000010);
	
	/* busy ? */
	lcd_busy();
	
	/* #5 function set 1 0 N F 0 0 
	   N=1 for 2 lines    N=0 for 1 line
	   F=0 for 5x8 chars  F=1 for 5x10 chars
	 */
	lcd_i_write(0b00101000);
	
	/* #6 display ON :  1 D C B  (Display, Cursor, Blink) */
	lcd_i_write(0b00001100);
	
	/* #7 display CLEAR */
	lcd_i_write(0b00000001);
	
	/* #8 entry mode */
	lcd_i_write(0b00000110);
	
	/* DDRAM address 0000 */
	lcd_i_write(0b10000000);
	
	return;
}	


void putrs_lcd(const rom char *buffer)
{
	while (*buffer) {
		lcd_d_write(*buffer);
		buffer++;
	}
}


void puts_lcd(char *buffer)
{
	while (*buffer) {
		lcd_d_write(*buffer);
		buffer++;
	}
}
	
	


char ntox(unsigned char nibble)
{
	nibble &= 0x0F;
	
	if (nibble < 10) {
		return ('0' + nibble);
	}
	nibble -= 10;
	return ('A' + nibble);	
}


char *btox(unsigned char x, char *str)
{
	str[0] = ntox((x & 0xF0) >> 4);
	str[1] = ntox(x & 0x0F);
	str[2] = 0;
	return str;
}	

char *itox(unsigned int x, char *str)
{
	str[0] = ntox((x & 0xF000) >> 12);
	str[1] = ntox((x & 0x0F00) >> 8);
	str[2] = ntox((x & 0x00F0) >> 4);
	str[3] = ntox(x & 0x000F);
	str[4] = 0;
	return str;	
}

char *ltox(unsigned long x, char *str)
{
	unsigned int y;
	
	y = (unsigned int)(x >> 16);
	itox(y, str);
	y = (unsigned int)x;
	itox(y, str + 4);
	
	return str;
}
